
/* Extended Module Player
 * Copyright (C) 1997-1998 Claudio Matsuoka and Hipolito Carraro Jr
 * $Id: mixer.c,v 1.3 1998/10/19 16:56:48 claudio Exp $
 *
 * This file is part of the Extended Module Player and is distributed
 * under the terms of the GNU General Public License. See doc/COPYING
 * for more information.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "mixer.h"

#define SMIX_NUMVOC     64        /* default number of softmixer voices */
#define SMIX_MAXRAT     48000        /* max sampling rate (Hz) */
#define SMIX_MINBPM     0x21        /* min BPM */
#define SMIX_RESMAX     (sizeof (int16))        /* max output resolution */

#define C4_NOTE         6947

#define FLAG_ITPT       0x01
#define FLAG_16_BITS    0x02
#define FLAG_STEREO     0x04
#define FLAG_BACKWARD   0x08
#define FLAG_ACTIVE     0x10
#define FIDX_FLAGMASK   (FLAG_ITPT | FLAG_16_BITS | FLAG_STEREO)

#define LIM_FT          12
#define LIM8_HI         127
#define LIM8_LO         -127
#define LIM12_HI        4095
#define LIM12_LO        -4096
#define LIM16_HI        32767
#define LIM16_LO        -32768

#define PATCH_OK        -1
#define OUT_SIZE        (15 * SMIX_MAXRAT * 2 * SMIX_RESMAX / SMIX_MINBPM / 6)
/* outsize = 15 * rate * mode[mn,st] * res[8,12,16,24b] / BPM_MIN / 6; */

static char **smix_buffer = NULL;        /* array of output buffers */
static int *smix_tbuffer = NULL;        /* temp buffer for 32 bit samples */
static int smix_numvoc;                /* default softmixer voices number */
int *tbuffer;                        /* same thing, aux ptr */
int Itpt_inc;                        /* interpolator increment */
int vol_right;                        /* right channel volume */
int vol_left;                        /* left channel/mono volume */

static int echo_msg;

extern int xmp_bpm;

void mix_mn8norm(int, char *, int, int);
void mix_mn8itpt(int, char *, int, int);
void mix_mn16norm(int, int16 *, int, int);
void mix_mn16itpt(int, int16 *, int, int);
void mix_st8norm(int, char *, int, int);
void mix_st8itpt(int, char *, int, int);
void mix_st16norm(int, int16 *, int, int);
void mix_st16itpt(int, int16 *, int, int);

static void out_su8norm(int, int *, char *, int);
static void out_su16norm(int, int *, int16 *, int);
static void out_u8ulaw(int, int *, char *, int);

/* Array index:

 * bit 0: 0=noninterpolated, 1=interpolated
 * bit 1: 0=8 bit, 1=16 bit
 * bit 2: 0=mono, 1=stereo
 */

static void (*mix_fn[]) () = {
   mix_mn8norm,
      mix_mn8itpt,
      mix_mn16norm,
      mix_mn16itpt,
      mix_st8norm,
      mix_st8itpt,
      mix_st16norm,
      mix_st16itpt
};

#define OUT_U8ULAW      0
#define OUT_SU8NORM     1
#define OUT_SU16NORM    2

static void (*out_fn[]) () = {
   out_u8ulaw,
      out_su8norm,
      out_su16norm
};

static int buffer_size;

int xmp_smix_is16bit(int outfmt)
{
   int type;

   type = !!(outfmt & (XFMT_U16_LE | XFMT_S16_LE | XFMT_U16_BE | XFMT_S16_BE));
   if (!type)
      type -= !(outfmt & (XFMT_U8 | XFMT_S8 | XFMT_MU_LAW));

   return type;                        /* -1 = wrong format, 0 = 8bit and 1 = 16bit output */
}


/* Handler 32bit samples to 8bit, signed or unsigned, mono or stereo output */
static void out_su8norm(int num, int *og, char *dt, int cod)
{
   int smp, lhi, llo, left, mask;

   mask = 0;
   left = LIM_FT + 8;
   switch (cod) {
   case XFMT_U8:
      mask = 0x80;
   case XFMT_S8:
      lhi = LIM8_HI;
      llo = LIM8_LO;
      break;
   default:
      return;
   }

   for (; num--; ++og, ++dt) {
      smp = *og >> left;
      *dt = smp > lhi ? lhi + mask : smp < llo ? llo + mask : smp + mask;
   }
}


/* Handler 32bit samples to 16bit, signed or unsigned, mono or stereo output */
static void out_su16norm(int num, int *og, int16 * dt, int cod)
{
   int smp, lhi, llo, left, mask;

   mask = 0;
   left = LIM_FT;
   switch (cod) {
   case XFMT_U16_LE:
   case XFMT_U16_BE:
      mask = 0x8000;
   case XFMT_S16_LE:
   case XFMT_S16_BE:
      lhi = LIM16_HI;
      llo = LIM16_LO;
      break;
   default:
      return;
   }

   for (; num--; ++og, ++dt) {
      smp = *og >> left;
      *dt = smp > lhi ? lhi + mask : smp < llo ? llo + mask : smp + mask;
   }
}


/* Handler 32bit samples to 8bit, unsigned ulaw, mono or stereo output */
static void out_u8ulaw(int num, int *og, char *dt, int cod)
{
   int smp, lhi, llo, left;

   if (cod != XFMT_MU_LAW)
      return;

   lhi = LIM12_HI;
   llo = LIM12_LO;
   left = LIM_FT + 4;

   for (; num--; ++og, ++dt) {
      smp = *og >> left;
      *dt = smp > lhi ? ulaw_encode(lhi) :
         smp < llo ? ulaw_encode(llo) : ulaw_encode(smp);
   }
}


/* Fill the output buffer calling one of the handlers. The buffer contains
 * sound for one tick (a PAL frame or 1/50s for standard vblank-timed mods)
 */
static int softmixer()
{
   int voc, t_cnt, p_cnt, dir;
   int lstart, lend, ticksize;
   struct voice_info *vi;
   struct patch_info *pi;

   int mode = xmp_ctl->flags & XMP_CTL_MONO ? 1 : 2;

   if (numvoc <= 0)
      return 0;

   ticksize = xmp_ctl->flags & XMP_CTL_MEDBPM ?
      xmp_ctl->freq * xmp_ctl->rrate * 33 / xmp_bpm / 12500 :
      xmp_ctl->freq * xmp_ctl->rrate / xmp_bpm / 100;

   buffer_size = ticksize * mode;

   if (smix_tbuffer)                /* Clear the buffer */
      memset(smix_tbuffer, 0, buffer_size * sizeof(int));

   for (voc = numvoc; voc--;) {
      vi = &voice_array[voc];
      if (vi->chn == -1)
         continue;
      if (!vi->period) {
         drv_resetvoice(voc);
         continue;
      }

      dir = 1;
      tbuffer = smix_tbuffer;
      pi = patch_array[vi->smp];
      vol_left = (vi->vol * (0x80 + vi->pan)) >> VOL_FT;
      vol_right = (vi->vol * (0x80 - vi->pan)) >> VOL_FT;
      Itpt_inc = ((long long) vi->pbase << SHIFT_FT) / vi->period;

      /* This is for bidirectional sample loops */
      if (vi->fidx & FLAG_BACKWARD) {
         Itpt_inc = -Itpt_inc;
         dir = 0;
      }

      /* Sample loop processing */
      lstart = pi->loop_start;
      lend = pi->loop_end;

      /* Offsets in samples, not bytes */
      if (vi->fidx & FLAG_16_BITS) {
         lstart >>= 1;
         lend >>= 1;
      }

      for (t_cnt = ticksize; t_cnt;) {
         /* How many samples we can write before the loop break or
          * sample end... */
         p_cnt = 1 + (((long long) (vi->end - vi->pos) << SHIFT_FT)
                      - vi->itpt) / Itpt_inc;

         if (dir) {
            if (vi->end < vi->pos)
               p_cnt = 0;
         }
         else if (vi->end > vi->pos)
            p_cnt = 0;

         /* ...inside the tick boundaries */
         if (p_cnt > t_cnt)
            p_cnt = t_cnt;

         t_cnt -= p_cnt;

         if (vi->vol)                /* Call the output handler */
            mix_fn[vi->fidx & FIDX_FLAGMASK] (p_cnt, vi->sptr, vi->pos - 1,
                                              vi->itpt + (1 << SHIFT_FT));
         tbuffer += p_cnt * mode;
         vi->itpt += Itpt_inc * p_cnt;
         vi->pos += vi->itpt >> SHIFT_FT;
         vi->itpt &= AND_FT;

         /* No more samples in this tick */
         if (!t_cnt)
            continue;
         /* Single shot sample */
         if (!(vi->fidx ^= vi->fxor) || lstart >= lend) {
            drv_resetvoice(voc);
            t_cnt = 0;
            continue;
         }

         if (!(vi->fidx & FLAG_BACKWARD || vi->fxor)) {
            vi->pos -= lend - lstart;
         }
         else {
            dir ^= 0x01;
            Itpt_inc = -Itpt_inc;
            vi->itpt += Itpt_inc;
            vi->pos += vi->itpt >> SHIFT_FT;
            vi->itpt &= AND_FT;
            vi->end = Itpt_inc > 0 ? lend : lstart;
         }
      }
   }

   ticksize *= mode;
   return ticksize << xmp_smix_is16bit(xmp_ctl->outfmt);
}

void *xmp_smix_buffer()
{
   static int buffer_switch = 0;
   void *smix_outptr;
   int act = 0;

   /* The mixer works with multiple buffers -- this is useless when
    * using multi-buffered sound output (e.g. OSS fragments) but
    * can be useful for DMA transfers in DOS.
    */
   smix_outptr = smix_buffer[buffer_switch++];
   if (buffer_switch >= xmp_ctl->numbuf)
      buffer_switch = 0;

   /* Transfer data from the 32-bit mixing buffer to the output buffer */

   switch (xmp_ctl->outfmt) {
   case XFMT_MU_LAW:
      act = OUT_U8ULAW;
      break;
   case XFMT_U8:
   case XFMT_S8:
      act = OUT_SU8NORM;
      break;
   case XFMT_U16_LE:
   case XFMT_S16_LE:
   case XFMT_U16_BE:
   case XFMT_S16_BE:
      act = OUT_SU16NORM;
      break;
   default:
      return smix_outptr;
   }
   out_fn[act] ((int) (buffer_size), smix_tbuffer, smix_outptr,
                xmp_ctl->outfmt);

   return smix_outptr;
}


int xmp_smix_init(struct xmp_control *ctrl)
{
   int count;

   smix_numvoc = SMIX_NUMVOC;

   if (smix_tbuffer)                /* if ready return now */
      return XMP_OK;

   if (ctrl->numbuf < 2)
      ctrl->numbuf = 2;

   if ((smix_tbuffer = Ecalloc(OUT_SIZE / 2, sizeof(int))) == NULL)
        return XMP_ERR_ALSTB;

   if ((smix_buffer = Ecalloc(ctrl->numbuf, sizeof(void *))) == NULL) {
      Efree(smix_tbuffer);
      smix_tbuffer = NULL;
      return XMP_ERR_ALSAR;
   }
   for (count = 0; count < ctrl->numbuf; count++) {
      if ((smix_buffer[count] = Ecalloc(1, OUT_SIZE)) == NULL) {
         while (count)
            Efree(smix_buffer[--count]);
         Efree(smix_buffer);
         smix_buffer = NULL;
         Efree(smix_tbuffer);
         smix_tbuffer = NULL;
         return XMP_ERR_ALSBK;
      }
   }

   extern_drv = 0;                /* active smix to output */
   return XMP_OK;                /* ok, it's ready now */
}


void xmp_smix_close()
{
   if (xmp_ctl->numbuf <= 0)
      return;
   while (xmp_ctl->numbuf--)
      Efree(smix_buffer[xmp_ctl->numbuf]);
   Efree(smix_buffer);
   Efree(smix_tbuffer);
   smix_buffer = NULL;
   smix_tbuffer = NULL;
   extern_drv = 1;                /* turn off smix to output, active extern drv */
}


static void smix_setpatch(int voc, int smp)
{
   struct voice_info *vi = &voice_array[voc];
   struct patch_info *pi = patch_array[smp];

   vi->smp = smp;
   vi->vol = vi->pos = vi->itpt = 0;
   vi->end = pi->len - 1;
   vi->freq = pi->base_freq;

   vi->note = PATCH_OK;

   vi->fidx = xmp_ctl->flags & XMP_CTL_ITPT ?
      FLAG_ITPT | FLAG_ACTIVE : FLAG_ACTIVE;

   if (~xmp_ctl->flags & XMP_CTL_MONO) {
      vi->fidx |= FLAG_STEREO;
      vi->pan = pi->panning;
   }
   else
      vi->pan = 0;

   vi->sptr = pi->data;
   if (TEST_FLAG(pi->mode, WAVE_16_BITS)) {
      vi->fidx |= FLAG_16_BITS;
      vi->end -= 1;
   }

   if (!TEST_FLAG(pi->mode, WAVE_LOOPING)) {
      vi->fxor = vi->fidx;
   }
   else {
      if (vi->end > pi->loop_end)
         vi->end = pi->loop_end;

      vi->fxor = FLAG_BACKWARD *
         TEST_FLAG(pi->mode, WAVE_BIDIR_LOOP);
   }

   if (vi->fidx & FLAG_16_BITS)
      vi->end >>= 1;
}


static void smix_setnote(int voc, int nte)
{
   struct voice_info *vi = &voice_array[voc];

   if (vi->note != PATCH_OK)
      smix_setpatch(voc, vi->smp);
   vi->period = note_to_period2(vi->note = nte, 0);
   nte = (long long) C4_NOTE *vi->freq / xmp_ctl->freq;

   vi->pbase = C4_FREQ * nte / patch_array[vi->smp]->base_note;
}


static void smix_voicepos(int voc, int index, int itpt)
{
   struct voice_info *vi = &voice_array[voc];
   struct patch_info *pi = patch_array[vi->smp];
   int lend;

   lend = pi->mode & WAVE_LOOPING ? pi->loop_end : pi->len - 1;

   if (vi->fidx & FLAG_16_BITS) {
      --lend;
      index >>= 1;
      lend >>= 1;
   }

   if (index < lend) {
      vi->itpt = itpt;
      vi->pos = index;
      if (vi->fidx & FLAG_BACKWARD) {
         vi->fidx ^= vi->fxor;
         vi->end = lend;
      }
   }
   else
      drv_resetvoice(voc);        /* messy data resets channel */
}


static void smix_setbend(int voc, int bend)
{
   voice_array[voc].period = note_to_period2(voice_array[voc].note, bend);
}


void xmp_smix_setvol(int voc, int vol)
{
   voice_array[voc].vol = vol;
}


void xmp_smix_setpan(int voc, int pan)
{
   if (~xmp_ctl->flags & XMP_CTL_MONO)
      voice_array[voc].pan = pan;
}


void xmp_smix_echoback(int msg)
{
   xmp_event_callback(echo_msg = msg);
}


int xmp_smix_getmsg()
{
   return echo_msg;
}


int xmp_smix_numvoices(int num)
{
   return num > smix_numvoc ? smix_numvoc : num;
}


int xmp_smix_writepatch(struct patch_info *patch)
{
   if (!patch)
      return XMP_OK;
   if (patch->len < 0)
      return XMP_ERR_TPTCH;

#ifdef WORDS_BIGENDIAN_TEST_
   if (patch->mode & WAVE_16_BITS)
      change_sex(patch->len, patch->data);
#endif
   if (patch->mode & WAVE_UNSIGNED)
      sig2uns(patch->len, patch->mode & WAVE_16_BITS, patch->data);

   return XMP_OK;
}
